﻿QWC = LibStub("AceAddon-3.0"):NewAddon("Quest Watch Config", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0");
local LSM = LibStub("LibSharedMedia-3.0");
local COLOR = "\124cff80ff50";
local COLOR_RED = "\124cffff3030";
local COLOR_WHITE = "\124cffffffff";
local TS_NUMB = 1; -- text style - number first
local TS_DESC = 2; -- text style - description first
local TC_USER = 1; -- text color - user defined
local TC_DIFF = 2; -- text color - quest difficulty
local questEvents = {
	["log"] = function(idx, questID)
		QuestLog_SetSelection(idx);
		ShowUIPanel(QuestLogFrame);
		QuestLog_Update();
	end,
	["map"] = function(idx, questID)
		WorldMap_OpenToQuest(questID);
	end,
	["link"] = function(idx, questID)
		if ChatFrameEditBox and ChatFrameEditBox:IsVisible() then
			ChatFrameEditBox:Insert(GetQuestLink(idx))
		else
			ChatEdit_InsertLink(GetQuestLink(idx))
		end
	end,
	["remove"] = function(idx, questID)
		RemoveQuestWatch(idx);
		QuestLog_Update();
	end,
};
local achievementEvents = {
	["log"] = function(aID)
		if ( not AchievementFrame ) then
			AchievementFrame_LoadUI();
		end
		if not AchievementFrame:IsShown() then
			AchievementFrame_ToggleAchievementFrame();
			AchievementFrame_SelectAchievement(aID);
		else
			if ( AchievementFrameAchievements.selection ~= aID ) then
				AchievementFrame_SelectAchievement(aID);
			else
				AchievementFrame_ToggleAchievementFrame();
			end
		end
	end,
	["map"] = function(aID)
		if ( not AchievementFrame ) then
			AchievementFrame_LoadUI();
		end
		if not AchievementFrame:IsShown() then
			AchievementFrame_ToggleAchievementFrame();
			AchievementFrame_SelectAchievement(aID);
		else
			if ( AchievementFrameAchievements.selection ~= aID ) then
				AchievementFrame_SelectAchievement(aID);
			else
				AchievementFrame_ToggleAchievementFrame();
			end
		end
	end,
	["link"] = function(aID)
		if ChatFrameEditBox and ChatFrameEditBox:IsVisible() then
			ChatFrameEditBox:Insert(GetAchievementLink(aID))
		else
			ChatEdit_InsertLink(GetAchievementLink(aID))
		end
	end,
	["remove"] = function(aID)
		RemoveTrackedAchievement(aID);
	end,
};
local noteEvents = {
	["log"] = function()
		QWCNoteFrame:Show();
	end,
	["map"] = function()
		QWCNoteFrame:Show();
	end,
	["link"] = function()
		QWCNoteFrame:Show();
	end,
	["remove"] = function()
		QWC.db.profile["note"] = nil;
		QWC:OnQuestLogChanged();
	end,
};
local L = LibStub("AceLocale-3.0"):GetLocale("QWC");
local bgTexture;
local texList;
local list = {
	quest = {},
	achievement = {},
};
local fsList; -- font string list
local qiList; -- quest item list
local qbList; -- quest button list
local pqList; -- popup quest list
local events = {
	["QUEST_LOG_UPDATE"] = "OnQuestUpdate",
	["TRACKED_ACHIEVEMENT_UPDATE"] = "UpdateAchievement",
	["RECEIVED_ACHIEVEMENT_LIST"] = "OnReceivedAchievementList",
	["QUEST_ACCEPTED"] = "OnQuestAccepted",
};
local hooks = {
	["AddQuestWatch"] = "OnQuestLogChanged",
	["RemoveQuestWatch"] = "OnQuestLogChanged",
	["AddTrackedAchievement"] = "AddAchievement",
	["RemoveTrackedAchievement"] = "RemoveAchievement",
	["WatchFrameAutoQuest_AddPopUp"] = "AddPopUp",
	["WatchFrameAutoQuest_ClearPopUp"] = "RemovePopUp",
};

local defaults = {
	profile = {
		edgeFile = "Blizzard Dialog",
		bgTexture = "Blizzard Parchment 2",
		tile = false,
		width = "250",
		height = "400",
		titleFont = "Friz Quadrata TT",
		objectiveFont = "Friz Quadrata TT",
		titleSize = "12",
		objectiveSize = "11",
		titleColor = { 1.0, 1.0, 0.0, 1.0 },
		objectiveColor = { 1.0, 1.0, 1.0, 1.0 },
		titleJustification = "LEFT",
		objectiveJustification = "LEFT",
		titleHighlight = { 0.75, 1.0, 0.5, 1.0},
		objectiveHighlight = { 0.8, 1.0, 0.8, 1.0 },
		titleInset = 11,
		objectiveInset = 8,
		verticalInset = 9,
		alpha = 1.0,
		lock = false,
		anchorX = 0,
		anchorY = 0,
		anchorPoint = "CENTER",
		insetTop = 4,
		insetBottom = 4,
		insetLeft = 4,
		insetRight = 4,
		edgeSize = "24",
		textStyle = TS_NUMB,
		clickable = true,
		map = "LeftButton",
		log = "RightButton",
		link = "LeftButton",
		remove = "RightButton",
		mapModifier = 1,
		logModifier = 1,
		linkModifier = 2,
		removeModifier = 2,
		textColor = TC_USER,
		textObjColor = TC_USER,
		note = nil,
	},
	char = {
		trackedAchievements = {},
		popupQuests = {},
	},
};

local options = {
    name = L["Quest Watch Config"],
    handler = QWC,
    type = "group",
    args = {
		show = {
			type = "toggle",
			guiHidden = true,
			name = L["Show Window"],
			desc = L["Show the Quest Watch window"],
			set = "Show",
			get = "Show",
		},
		hide = {
			type = "toggle",
			guiHidden = true,
			name = L["Hide Window"],
			desc = L["Hide the Quest Watch window"],
			set = "Hide",
			get = "Hide",
		},
		config = {
			type = "execute",
			guiHidden = true,
			name = L["Configure"],
			desc = L["Open the configuration dialog"],
			func = function()
				InterfaceOptionsFrame_OpenToCategory(QWC.optFrame.Profiles);
				InterfaceOptionsFrame_OpenToCategory(QWC.optFrame);
			end,
		},
		lock = {
			type = "toggle",
			order = 0,
			name = L["Lock Window"],
			desc = L["Lock the window in place"],
			set = "Lock",
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		note = {
			type = "input",
			guiHidden = true,
			name = "Note",
			desc = "Set a note to appear at the top of the frame.".."\n".."Type "..COLOR_RED.."/qwc note clear"..COLOR_WHITE.." to remove the note.",
			set = function(info, value)
				if value == "" then
					QWCNoteFrame:Show();
				else
					QWC.db.profile[info[1]] = value;
					QWC:OnQuestLogChanged();
				end
				if QWC.db.profile[info[1]] then
					local title, body = QWC:SplitNote();
					QWCNoteEdit:SetText(title);
					QWCNoteEditBody:SetText(body);
				end
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		bgTexture = {
			type = "select",
			cmdHidden = true,
			order = 3,
			width = "double",
			name = L["Background"],
			desc = L["Set the window background"],
			values = LSM:HashTable("background"),
			dialogControl = "LSM30_Background",
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:ApplyBackground();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		tile = {
			type = "toggle",
			cmdHidden = true,
			order = 4,
			width = "half",
			name = L["Tile"],
			desc = L["Tile the background"].."\n"..L["Requires reload!"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				local size = QWC.db.profile["edgeSize"];
				if size == "0" then
					size = -1;
				end
				QWFrame:SetBackdrop({
					edgeFile = LSM:HashTable("border")[QWC.db.profile["edgeFile"]],
					edgeSize = size,
					tile = QWC.db.profile[info[1]],
				});
				QWC:ApplyBackground();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		alpha = {
			type = "range",
			cmdHidden = true,
			order = 5,
			min = 0.0,
			max = 1.0,
			isPercent = true,
			name = L["Background Visibility"],
			desc = L["Set the visibility of the background"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWCBGTEX:SetAlpha(QWC.db.profile[info[1]]);
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		edgeFile = {
			type = "select",
			cmdHidden = true,
			order = 6,
			width = "double",
			name = L["Border"],
			desc = L["Set the window border"],
			values = LSM:HashTable("border"),
			dialogControl = "LSM30_Border",
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				local size = QWC.db.profile["edgeSize"];
				if size == "0" then
					size = -1;
				end
				QWFrame:SetBackdrop({
					edgeFile = LSM:HashTable("border")[QWC.db.profile[info[1]]],
					edgeSize = size,
					tile = QWC.db.profile["tile"],
				});
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		width = {
			type = "range",
			cmdHidden = true,
			order = 1,
			min = 0,
			max = 4096,
			softMin = 32,
			softMax = 512,
			name = L["Width"],
			desc = L["Set the window width."],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWFrame:SetWidth(QWC.db.profile[info[1]]);
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		height = {
			type = "range",
			cmdHidden = true,
			order = 2,
			min = 0,
			max = 4096,
			softMin = 32,
			softMax = 512,
			name = L["Height"],
			desc = L["Set the window height."],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWFrame:SetHeight(QWC.db.profile[info[1]]);
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		titleFont = {
			type = "select",
			cmdHidden = true,
			order = 7,
			width = "double",
			name = L["Title Font"],
			desc = L["Set the title font"],
			values = LSM:HashTable("font"),
			dialogControl = "LSM30_Font",
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				for _, font in pairs(QWC.titleFonts) do
					font:SetFont(LSM:HashTable("font")[QWC.db.profile[info[1]]], QWC.db.profile["titleSize"]);
				end
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		titleSize = {
			type = "input",
			cmdHidden = true,
			order = 8,
			name = L["Title Size"],
			desc = L["Size of the title font"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				for _, font in pairs(QWC.titleFonts) do
					font:SetFont(LSM:HashTable("font")[QWC.db.profile["titleFont"]], QWC.db.profile[info[1]]);
				end
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		objectiveSize = {
			type = "input",
			cmdHidden = true,
			order = 10,
			name = L["Objective Size"],
			desc = L["Size of the objective font"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				for _, font in pairs(QWC.textFonts) do
					font:SetFont(LSM:HashTable("font")[QWC.db.profile["objectiveFont"]], QWC.db.profile[info[1]]);
				end
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		objectiveFont = {
			type = "select",
			cmdHidden = true,
			order = 9,
			width = "double",
			name = L["Objective Font"],
			desc = L["Set the objective font"],
			values = LSM:HashTable("font"),
			dialogControl = "LSM30_Font",
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				for _, font in pairs(QWC.textFonts) do
					font:SetFont(LSM:HashTable("font")[QWC.db.profile[info[1]]], QWC.db.profile["objectiveSize"]);
				end
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		insetTop = {
			type = "range",
			cmdHidden = true,
			order = 11,
			min = -512.0,
			max = 512.0,
			softMin = -15.0,
			softMax = 15.0,
			name = L["Background Inset (Top)"],
			desc = L["Set the top position of the background"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:ApplyBackground();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		insetBottom = {
			type = "range",
			cmdHidden = true,
			order = 12,
			min = -512.0,
			max = 512.0,
			softMin = -15.0,
			softMax = 15.0,
			name = L["Background Inset (Bottom)"],
			desc = L["Set the bottom position of the background"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:ApplyBackground();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		insetLeft = {
			type = "range",
			cmdHidden = true,
			order = 13,
			min = -512.0,
			max = 512.0,
			softMin = -15.0,
			softMax = 15.0,
			name = L["Background Inset (Left)"],
			desc = L["Set the left position of the background"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:ApplyBackground();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		insetRight = {
			type = "range",
			cmdHidden = true,
			order = 14,
			min = -512.0,
			max = 512.0,
			softMin = -15.0,
			softMax = 15.0,
			name = L["Background Inset (Right)"],
			desc = L["Set the right position of the background"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:ApplyBackground();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		edgeSize = {
			type = "input",
			cmdHidden = true,
			order = 15,
			name = L["Border Size"],
			desc = L["Set the size of the border"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				local size = QWC.db.profile[info[1]];
				if size == "0" then
					size = -1;
				end
				QWFrame:SetBackdrop({
					edgeFile = LSM:HashTable("border")[QWC.db.profile["edgeFile"]],
					edgeSize = size,
					tile = QWC.db.profile["tile"],
				});
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		titleColor = {
			type = "color",
			cmdHidden = true,
			order = 16,
			hasAlpha = true,
			name = L["Title Color"],
			desc = L["Color of the title font"],
			set = function(info, r, g, b, a)
				QWC.db.profile[info[1]] = { r, g, b, a };
				QWCTitle:SetTextColor(
					QWC.db.profile[info[1]][1],
					QWC.db.profile[info[1]][2],
					QWC.db.profile[info[1]][3],
					QWC.db.profile[info[1]][4]
				);
			end,
			get = function(info)
				return QWC.db.profile[info[1]][1], QWC.db.profile[info[1]][2], 
					QWC.db.profile[info[1]][3], QWC.db.profile[info[1]][4];
			end,
		},
		objectiveColor = {
			type = "color",
			cmdHidden = true,
			order = 17,
			hasAlpha = true,
			name = L["Objective Color"],
			desc = L["Color of the objective font"],
			set = function(info, r, g, b, a)
				QWC.db.profile[info[1]] = { r, g, b, a };
				QWCText:SetTextColor(
					QWC.db.profile[info[1]][1],
					QWC.db.profile[info[1]][2],
					QWC.db.profile[info[1]][3],
					QWC.db.profile[info[1]][4]
				);
			end,
			get = function(info)
				return QWC.db.profile[info[1]][1], QWC.db.profile[info[1]][2], 
					QWC.db.profile[info[1]][3], QWC.db.profile[info[1]][4];
			end,
		},
		titleJustification = {
			type = "select",
			cmdHidden = true,
			order = 18,
			values = {
				["LEFT"] = L["Left"],
				["CENTER"] = L["Center"],
				["RIGHT"] = L["Right"],
			},
			name = L["Title Justification"],
			desc = L["Set the title text justification"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				for _, font in pairs(QWC.titleFonts) do
					font:SetJustifyH(QWC.db.profile[info[1]]);
				end
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		objectiveJustification = {
			type = "select",
			cmdHidden = true,
			order = 19,
			values = {
				["LEFT"] = L["Left"],
				["CENTER"] = L["Center"],
				["RIGHT"] = L["Right"],
			},
			name = L["Objective Justification"],
			desc = L["Set the objective text justification"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				for _, font in pairs(QWC.textFonts) do
					font:SetJustifyH(QWC.db.profile[info[1]]);
				end
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		titleInset = {
			type = "range",
			cmdHidden = true,
			order = 20,
			min = -512.0,
			max = 512.0,
			softMin = 0.0,
			softMax = 32.0,
			name = L["Title Inset"],
			desc = L["Set the amount to indent the title"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		objectiveInset = {
			type = "range",
			cmdHidden = true,
			order = 21,
			min = -512.0,
			max = 512.0,
			softMin = 0.0,
			softMax = 32.0,
			name = L["Objective Inset"],
			desc = L["Set the amount to indent the objectives"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		verticalInset = {
			type = "range",
			cmdHidden = true,
			order = 22,
			min = -512.0,
			max = 512.0,
			softMin = 0.0,
			softMax = 32.0,
			name = L["Vertical Inset"],
			desc = L["Set the spacing at the top and bottom of the frame"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		textStyle = {
			type = "select",
			cmdHidden = true,
			order = 23,
			values = {
				[TS_NUMB] = L["Numbers First"],
				[TS_DESC] = L["Numbers Last"],
			},
			name = L["Text Style"],
			desc = L["Set the objective text justification"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		clickable = {
			type = "toggle",
			cmdHidden = true,
			order = 24,
			name = L["Clickable Quests"],
			desc = L["Allows you to click on quests to open the quest log and world map"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		titleHighlight = {
			type = "color",
			cmdHidden = true,
			order = 25,
			hasAlpha = true,
			name = L["Title Highlight"],
			desc = L["Color of the title font when the mouse is over the text"],
			set = function(info, r, g, b, a)
				QWC.db.profile[info[1]] = { r, g, b, a };
				QWCTitleHighlight:SetTextColor(
					QWC.db.profile[info[1]][1],
					QWC.db.profile[info[1]][2],
					QWC.db.profile[info[1]][3],
					QWC.db.profile[info[1]][4]
				);
			end,
			get = function(info)
				return QWC.db.profile[info[1]][1], QWC.db.profile[info[1]][2], 
					QWC.db.profile[info[1]][3], QWC.db.profile[info[1]][4];
			end,
		},
		objectiveHighlight = {
			type = "color",
			cmdHidden = true,
			order = 26,
			hasAlpha = true,
			name = L["Objective Highlight"],
			desc = L["Color of the objective font when the mouse is over the text"],
			set = function(info, r, g, b, a)
				QWC.db.profile[info[1]] = { r, g, b, a };
				QWCTextHighlight:SetTextColor(
					QWC.db.profile[info[1]][1],
					QWC.db.profile[info[1]][2],
					QWC.db.profile[info[1]][3],
					QWC.db.profile[info[1]][4]
				);
			end,
			get = function(info)
				return QWC.db.profile[info[1]][1], QWC.db.profile[info[1]][2], 
					QWC.db.profile[info[1]][3], QWC.db.profile[info[1]][4];
			end,
		},
		map = {
			type = "select",
			cmdHidden = true,
			order = 27,
			name = L["Map Mouse Click"],
			desc = L["Set the mouse button that will open the world map to the selected quest"],
			values = {
				["LeftButton"] = L["Left Button"],
				["RightButton"] = L["Right Button"],
				["MiddleButton"] = L["Middle Button"],
				["Button4"] = L["Button 4"],
				["Button5"] = L["Button 5"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		mapModifier = {
			type = "select",
			cmdHidden = true,
			order = 28,
			name = L["Map Mouse Modifier"],
			desc = L["Set the modifier to accompany the mouse click."],
			values = {
				L["None"],
				L["Shift"],
				L["Alt"],
				L["Control"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		log = {
			type = "select",
			cmdHidden = true,
			order = 29,
			name = L["Log Mouse Click"],
			desc = L["Set the mouse button that will open the quest log to the selected quest"],
			values = {
				["LeftButton"] = L["Left Button"],
				["RightButton"] = L["Right Button"],
				["MiddleButton"] = L["Middle Button"],
				["Button4"] = L["Button 4"],
				["Button5"] = L["Button 5"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		logModifier = {
			type = "select",
			cmdHidden = true,
			order = 30,
			name = L["Log Mouse Modifier"],
			desc = L["Set the modifier to accompany the mouse click."],
			values = {
				L["None"],
				L["Shift"],
				L["Alt"],
				L["Control"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		link = {
			type = "select",
			cmdHidden = true,
			order = 31,
			name = L["Link Mouse Click"],
			desc = L["Set the mouse button that will create a link to the quest or achievement"],
			values = {
				["LeftButton"] = L["Left Button"],
				["RightButton"] = L["Right Button"],
				["MiddleButton"] = L["Middle Button"],
				["Button4"] = L["Button 4"],
				["Button5"] = L["Button 5"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		linkModifier = {
			type = "select",
			cmdHidden = true,
			order = 32,
			name = L["Link Mouse Modifier"],
			desc = L["Set the modifier to accompany the mouse click."],
			values = {
				L["None"],
				L["Shift"],
				L["Alt"],
				L["Control"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		remove = {
			type = "select",
			cmdHidden = true,
			order = 33,
			name = L["Remove Mouse Click"],
			desc = L["Set the mouse button that will stop tracking the quest or achievement"],
			values = {
				["LeftButton"] = L["Left Button"],
				["RightButton"] = L["Right Button"],
				["MiddleButton"] = L["Middle Button"],
				["Button4"] = L["Button 4"],
				["Button5"] = L["Button 5"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		removeModifier = {
			type = "select",
			cmdHidden = true,
			order = 34,
			name = L["Remove Mouse Modifier"],
			desc = L["Set the modifier to accompany the mouse click."],
			values = {
				L["None"],
				L["Shift"],
				L["Alt"],
				L["Control"],
			},
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		textColor = {
			type = "select",
			cmdHidden = true,
			order = 35,
			values = {
				[TC_USER] = L["User Defined"],
				[TC_DIFF] = L["Quest Difficulty"],
			},
			name = L["Title Text Color"],
			desc = L["Set the color style for title text"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
		textObjColor = {
			type = "select",
			cmdHidden = true,
			order = 36,
			values = {
				[TC_USER] = L["User Defined"],
				[TC_DIFF] = L["Quest Difficulty"],
			},
			name = L["Objective Text Color"],
			desc = L["Set the color style for objective text"],
			set = function(info, value)
				QWC.db.profile[info[1]] = value;
				QWC:OnQuestLogChanged();
			end,
			get = function(info)
				return QWC.db.profile[info[1]];
			end,
		},
	},
}

function QWC:ApplyBackground()
	bgTexture = bgTexture or QWFrame:CreateTexture("QWCBGTEX", "BACKGROUND");
	bgTexture:SetTexture(nil);
	bgTexture:SetHorizTile(self.db.profile["tile"]);
	bgTexture:SetVertTile(self.db.profile["tile"]);
	bgTexture:SetTexture(LSM:HashTable("background")[self.db.profile["bgTexture"]], self.db.profile["tile"]);
	bgTexture:ClearAllPoints();
	--bgTexture:SetAllPoints(QWFrame);
	bgTexture:SetPoint("TOPLEFT", QWFrame, "TOPLEFT", self.db.profile["insetLeft"], -self.db.profile["insetTop"]);
	bgTexture:SetPoint("BOTTOMRIGHT", QWFrame, "BOTTOMRIGHT", -self.db.profile["insetRight"], self.db.profile["insetBottom"]);
end

function QWC:InitTexList()
	texList = texList or {}
	for i,j in pairs(LSM:HashTable("border")) do
		texList[i] = QWFrame:CreateTexture(nil, "BORDER");
		texList[i]:SetTexture(j, true);
	end
end

function QWC:Lock(info, input)
	self.db.profile["lock"] = input
	QWFrame:EnableMouse(not input);
	if input then
		if QWCMover ~= nil then
			QWCMover:Hide();
		end
	else
		if QWCMover == nil then
			local frame = CreateFrame("Frame", "QWCMover", UIParent);
			frame:SetMovable(true);
			frame:EnableMouse(true);
			frame:SetScript("OnMouseDown", function(self, button)
				if button == "LeftButton" and not self.isMoving then
					self:StartMoving();
					self.isMoving = true;
				end
			end);
			frame:SetScript("OnMouseUp", function(self, button)
				if button == "LeftButton" and self.isMoving then
					self:StopMovingOrSizing();
					self.isMoving = false;
					local f1, f2, f3, f4, f5 = self:GetPoint(1);
					QWC.db.profile["anchorPoint"] = f1;
					QWC.db.profile["anchorX"] = f4;
					QWC.db.profile["anchorY"] = f5;
				end
			end);
			frame:SetScript("OnHide", function(self)
				if ( self.isMoving ) then
					self:StopMovingOrSizing();
					self.isMoving = false;
				end
			end);
			_, _, _, x, y = QWFrame:GetPoint(1);
			frame:SetFrameStrata("TOOLTIP");
			frame:SetPoint(self.db.profile["anchorPoint"], UIParent, self.db.profile["anchorPoint"], self.db.profile["anchorX"], self.db.profile["anchorY"]);
			frame:SetWidth(self.db.profile["width"]);
			frame:SetHeight(self.db.profile["height"]);
			QWFrame:SetParent(frame);
			QWFrame:ClearAllPoints();
			QWFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0);
			QWFrame:SetParent(UIParent);
			local tex = frame:CreateTexture("ARTWORK");
			tex:SetAllPoints();
			tex:SetTexture(0.5, 0.0, 0);
			tex:SetAlpha(0.5);
			local fs = frame:CreateFontString("QWCMoverText", "ARTWORK");
			fs:SetPoint("CENTER", frame, "CENTER");
			fs:SetFontObject("QWCMove");
			fs:SetText("Drag to move. Type \n/qwc lock\n to lock the frame in place.");
		else
			QWCMover:SetWidth(self.db.profile["width"]);
			QWCMover:SetHeight(self.db.profile["height"]);
			QWCMover:Show();
			QWFrame:SetParent(QWCMover);
			QWFrame:ClearAllPoints();
			QWFrame:SetPoint("TOPLEFT", QWCMover, "TOPLEFT", 0, 0);
			QWFrame:SetParent(UIParent);
		end
	end
end

function QWC:RefreshConfig()
	QWFrame:SetWidth(self.db.profile["width"]);
	QWFrame:SetHeight(self.db.profile["height"]);

	if QWC.db.profile["note"] then
		local title, body = QWC:SplitNote();
		QWCNoteEdit:SetText(title);
		QWCNoteEditBody:SetText(body);
	end
	
	QWCTitle:SetTextColor(
		QWC.db.profile["titleColor"][1],
		QWC.db.profile["titleColor"][2],
		QWC.db.profile["titleColor"][3],
		QWC.db.profile["titleColor"][4]
	);
	QWCText:SetTextColor(
		QWC.db.profile["objectiveColor"][1],
		QWC.db.profile["objectiveColor"][2],
		QWC.db.profile["objectiveColor"][3],
		QWC.db.profile["objectiveColor"][4]
	);
	QWCTitleHighlight:SetTextColor(
		QWC.db.profile["titleHighlight"][1],
		QWC.db.profile["titleHighlight"][2],
		QWC.db.profile["titleHighlight"][3],
		QWC.db.profile["titleHighlight"][4]
	);
	QWCTextHighlight:SetTextColor(
		QWC.db.profile["objectiveHighlight"][1],
		QWC.db.profile["objectiveHighlight"][2],
		QWC.db.profile["objectiveHighlight"][3],
		QWC.db.profile["objectiveHighlight"][4]
	);
	
	local size = self.db.profile["edgeSize"];
	if size == "0" then
		size = -1;
	end
	QWFrame:SetBackdrop({
		edgeFile = LSM:HashTable("border")[self.db.profile["edgeFile"]],
		edgeSize = size,
		tile = self.db.profile["tile"],
	});
	QWC:ApplyBackground();
	
	for _, font in pairs(QWC.titleFonts) do
		font:SetFont(LSM:HashTable("font")[self.db.profile["titleFont"]], self.db.profile["titleSize"]);
	end
	
	for _, font in pairs(QWC.textFonts) do
		font:SetFont(LSM:HashTable("font")[self.db.profile["objectiveFont"]], self.db.profile["objectiveSize"]);
	end
	
	QWCBGTEX:SetAlpha(self.db.profile["alpha"]);
	
	if WatchFrame:IsShown() then
		WatchFrame:Hide();
	end
	
	QWFrame:ClearAllPoints();
	QWFrame:SetPoint(self.db.profile["anchorPoint"], UIParent, self.db.profile["anchorPoint"], self.db.profile["anchorX"], self.db.profile["anchorY"]);
	
	QWC:Lock(nil, not self.db.profile["lock"]);
	QWC:Lock(nil, not self.db.profile["lock"]);
	
	QWCMover:SetPoint(self.db.profile["anchorPoint"], UIParent, self.db.profile["anchorPoint"], self.db.profile["anchorX"], self.db.profile["anchorY"]);
end

function QWC:SharedMedia_Registered(event, lib, entry)
	if lib == "border" then
		texList[entry] = QWFrame:CreateTexture(nil, "BORDER");
		texList[entry]:SetTexture(LSM:HashTable("border")[entry]);
		if entry == self.db.profile["edgeFile"] then
			QWC:RefreshConfig();
		end
	end
	if lib == "background" and entry == self.db.profile["bgTexture"] then
		QWC:RefreshConfig();
	end
end

function QWC:OnInitialize()
	LSM.RegisterCallback(QWC, "LibSharedMedia_Registered", "SharedMedia_Registered");
	LSM.RegisterCallback(QWC, "LibSharedMedia_Registered", "SharedMedia_Registered");

	QWC:InitTexList();
	
	QWC.titleFonts = {
		[1] = QWCTitle,
		[2] = QWCDiff_Trivial,
		[3] = QWCDiff_Standard,
		[4] = QWCDiff_Difficult,
		[5] = QWCDiff_VeryDifficult,
		[6] = QWCDiff_Impossible,
		[7] = QWCDiff_TrivialHighlight,
		[8] = QWCDiff_StandardHighlight,
		[9] = QWCDiff_DifficultHighlight,
		[10] = QWCDiff_VeryDifficultHighlight,
		[11] = QWCDiff_ImpossibleHighlight,
	};
	QWC.textFonts = {
		[1] = QWCText,
		[2] = QWCODiff_Trivial,
		[3] = QWCODiff_Standard,
		[4] = QWCODiff_Difficult,
		[5] = QWCODiff_VeryDifficult,
		[6] = QWCODiff_Impossible,
		[7] = QWCODiff_TrivialHighlight,
		[8] = QWCODiff_StandardHighlight,
		[9] = QWCODiff_DifficultHighlight,
		[10] = QWCODiff_VeryDifficultHighlight,
		[11] = QWCODiff_ImpossibleHighlight,
	};
	
	self.db = LibStub("AceDB-3.0"):New("QWCDB", defaults);
	
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
	
	local AC = LibStub("AceConfig-3.0");
	local ACD = LibStub("AceConfigDialog-3.0");
	local ACR = LibStub("AceConfigRegistry-3.0");
	
	for event, method in pairs(events) do
		QWC:RegisterEvent(event, method);
	end
	for hook, method in pairs(hooks) do
		QWC:SecureHook(hook, method);
	end
	
	AC:RegisterOptionsTable("Quest Watch Config", options, {"questwatchconfig", "qwc"});
	ACR:RegisterOptionsTable("QWCProfiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db));
	
	QWC.optFrame = ACD:AddToBlizOptions("Quest Watch Config", L["Quest Watch Config"]);
	QWC.optFrame.Profiles = ACD:AddToBlizOptions("QWCProfiles", "Profiles", "Quest Watch Config");
	
	QWC:RefreshConfig();
	QWC:OnQuestLogChanged();
	
	print(COLOR..L["Quest Watch Config Loaded Successfully! Type "]..COLOR_RED.."/qwc config"..COLOR..L[" to configure."]);
end

function pairsByKeys (t, f)
	local a = {}
	-- build temporary table of the keys
	for n in pairs (t) do 
		table.insert (a, n) 
	end
	table.sort (a, f)  -- sort using supplied function, if any
	local i = 0        -- iterator variable
	return function () -- iterator function
		i = i + 1
		return a[i], t[a[i]]
	end  -- iterator function
end -- pairsByKeys

function QWC:Test(info, input)
	QWCNoteEdit = QWCNoteEdit or CreateFrame("EDITBOX", "QWCNoteEdit", UIParent);
	QWCNoteEdit:ClearFocus();
end

function QWC:AddAchievement(achievementId)
	local _; -- used to ignore certain return values
	local info = self.db.char["trackedAchievements"][achievementId] or {};
	_, info.name, _, info.complete, _, _, _, info.desc = GetAchievementInfo(achievementId);
	info.objCount = GetAchievementNumCriteria(achievementId);
	for idx = 1, info.objCount do
		info["obj"..idx] = info["obj"..idx] or {};
		info["obj"..idx].objDesc, _, info["obj"..idx].objComplete, info["obj"..idx].quantity, info["obj"..idx].reqQuantity = GetAchievementCriteriaInfo(achievementId, idx);
	end
	tinsert(self.db.char["trackedAchievements"], achievementId, info);
	QWC:OnQuestLogChanged();
end

function QWC:RemoveAchievement(achievementId)
	self.db.char["trackedAchievements"][achievementId] = nil;
	QWC:OnQuestLogChanged();
end

function QWC:AddPopUp(questId, type)
	local _; -- used to ignore certain return values
	local info = self.db.char["popupQuests"][questId] or {};
	if info == true then
		info = {};
	end
	info.name, _, _, _, _, _, info.complete = GetQuestLogTitle(GetQuestLogIndexByID(questId));
	info.type = type;
	tinsert(self.db.char["popupQuests"], questId, info);
	QWC:OnQuestLogChanged();
end

function QWC:RemovePopUp(questId)
	if self.db.char["popupQuests"][questId].type == "COMPLETE" then
		self.db.char["popupQuests"][questId] = nil;
		self.db.char["popupQuests"][questId] = true;
	else
		self.db.char["popupQuests"][questId] = nil;
	end
	QWC:OnQuestLogChanged();
end

function QWC:Split(str, separator)
	if type( str ) ~= "string" or type( separator ) ~= "string" then
		return
	end
	local t = {}
	local fpat = "(.-)"..separator
	local last_end = 1
	local s, e, cap = str:find(fpat, 1)
	while s do
		if s ~= 1 or cap ~= "" then
			tinsert(t,cap)
		end
		last_end = e + 1
		s, e, cap = str:find(fpat, last_end)
	end
	if last_end <= #str then
		cap = str:sub(last_end)
		tinsert(t, cap)
	end
	return t
end

function QWC:Modifier(value)
	if value == 2 then
		return IsShiftKeyDown();
	elseif value == 3 then
		return IsAltKeyDown();
	elseif value == 4 then
		return IsControlKeyDown();
	end
	return not (IsShiftKeyDown() or IsAltKeyDown() or IsControlKeyDown());
end

function QWC:OnReceivedAchievementList(...)
	for achievementId, info in pairs(self.db.char["trackedAchievements"]) do
		QWC:AddAchievement(achievementId);
	end
end

function QWC:UpdateAchievement(source, achievementId)
	if list.achievement[achievementId] ~= nil then
		local _; -- used to ignore certain return values
		local info = self.db.char["trackedAchievements"][achievementId] or {};
		_, info.name, _, info.complete, _, _, _, info.desc = GetAchievementInfo(achievementId);
		info.objCount = GetAchievementNumCriteria(achievementId);
		for idx = 1, info.objCount do
			info["obj"..idx] = info["obj"..idx] or {};
			info["obj"..idx].objDesc, _, info["obj"..idx].objComplete, info["obj"..idx].quantity, info["obj"..idx].reqQuantity = GetAchievementCriteriaInfo(achievementId, idx);
		end
		tinsert(self.db.char["trackedAchievements"], achievementId, info);
		if info.complete == 1 or info.complete == true then
			QWC:OnQuestLogChanged();
			return;
		end
		if info.objCount == 0 then
			fs:SetText(info.desc);
		else
			for idx, fs in pairs(list.achievement[achievementId].child) do
				if info["obj"..idx].objComplete ~= true then
					if info["obj"..idx].reqQuantity > 1 then -- append ": X/Y" to match quest layout
						if self.db.profile["textStyle"] == TS_NUMB then
							fs:SetText(info["obj"..idx].quantity.."/"..info["obj"..idx].reqQuantity.." "..info["obj"..idx].objDesc);
						elseif self.db.profile["textStyle"] == TS_DESC then
							fs:SetText(info["obj"..idx].objDesc..": "..info["obj"..idx].quantity.."/"..info["obj"..idx].reqQuantity);
						end
					else
						fs:SetText(info["obj"..idx].objDesc);
					end
				else
					QWC:OnQuestLogChanged();
					return;
				end
			end
		end
	end
end

function QWC:OnQuestAccepted(source, logID)
	local _, _, _, _, _, _, complete, _, questID = GetQuestLogTitle(logID);
	if complete == nil or complete == false then
		self.db.char["popupQuests"][questID] = nil;
	end
end

function QWC:OnQuestUpdate(...)
	if list.quest == nil then
		return;
	end
	local _, numQuests = GetNumQuestLogEntries();
	if #list.quest ~= numQuests then
		QWC:OnQuestLogChanged();
		return;
	end
	for idx, qlist in pairs(list.quest) do
		local _, _, _, _, _, _, complete = GetQuestLogTitle(idx);
		if complete == 1 or complete == true then
			QWC:OnQuestLogChanged();
			return;
		end
		for jdx, fs in pairs(qlist.child) do
			local info, qType, done = GetQuestLogLeaderBoard(jdx, idx);
			if info ~= nil and done ~= 1 or qType == "log" then
				if self.db.profile["textStyle"] == TS_NUMB then
					local newInfo = QWC:Split(info, ": ");
					local fixedDesc = "";
					for kdx = 1, (#newInfo - 1) do
						fixedDesc = fixedDesc..newInfo[kdx];
					end
					fs:SetText(newInfo[#newInfo].." "..fixedDesc);
				elseif self.db.profile["textStyle"] == TS_DESC then
					fs:SetText(info);
				end
			else
				QWC:OnQuestLogChanged();
				return;
			end
		end
	end
end

function QWC:GetQuestDiffFont(level, isObjective)
	local levelDiff = level - UnitLevel("player");
	if isObjective == nil or isObjective == false then
		if ( levelDiff >= 5 ) then
			return "QWCDiff_Impossible";
		elseif ( levelDiff >= 3 ) then
			return "QWCDiff_VeryDifficult";
		elseif ( levelDiff >= -2 ) then
			return "QWCDiff_Difficult";
		elseif ( -levelDiff <= GetQuestGreenRange() ) then
			return "QWCDiff_Standard";
		else
			return "QWCDiff_Trivial";
		end
	else
		if ( levelDiff >= 5 ) then
			return "QWCODiff_Impossible";
		elseif ( levelDiff >= 3 ) then
			return "QWCODiff_VeryDifficult";
		elseif ( levelDiff >= -2 ) then
			return "QWCODiff_Difficult";
		elseif ( -levelDiff <= GetQuestGreenRange() ) then
			return "QWCODiff_Standard";
		else
			return "QWCODiff_Trivial";
		end
	end
end

function QWC:GetQuestDiffFontHighlight(level, isObjective)
	local levelDiff = level - UnitLevel("player");
	if isObjective == nil or isObjective == false then
		if ( levelDiff >= 5 ) then
			return "QWCDiff_ImpossibleHighlight";
		elseif ( levelDiff >= 3 ) then
			return "QWCDiff_VeryDifficultHighlight";
		elseif ( levelDiff >= -2 ) then
			return "QWCDiff_DifficultHighlight";
		elseif ( -levelDiff <= GetQuestGreenRange() ) then
			return "QWCDiff_StandardHighlight";
		else
			return "QWCDiff_TrivialHighlight";
		end
	else
		if ( levelDiff >= 5 ) then
			return "QWCODiff_ImpossibleHighlight";
		elseif ( levelDiff >= 3 ) then
			return "QWCODiff_VeryDifficultHighlight";
		elseif ( levelDiff >= -2 ) then
			return "QWCODiff_DifficultHighlight";
		elseif ( -levelDiff <= GetQuestGreenRange() ) then
			return "QWCODiff_StandardHighlight";
		else
			return "QWCODiff_TrivialHighlight";
		end
	end
end

function QWC:IsTitleFont(fontName)
	for _, font in pairs(QWC.titleFonts) do
		if font:GetName() == fontName then
			return true;
		end
	end
	return false;
end

function QWC:SplitNote()
	local data = QWC:Split(QWC.db.profile["note"], "\"");
	return data[1], data[3];
end

function QWC:GetLogClickText()
	local retVal = "";
	if self.db.profile["logModifier"] == 2 then
		retVal = L["Shift"].."-";
	elseif self.db.profile["logModifier"] == 3 then
		retVal = L["Alt"].."-";
	elseif self.db.profile["logModifier"] == 4 then
		retVal = L["Control"].."-";
	end
	if self.db.profile["log"] == "LeftButton" then
		retVal = retVal..L["Left click to complete"];
	elseif self.db.profile["log"] == "RightButton" then
		retVal = retVal..L["Right click to complete"];
	elseif self.db.profile["log"] == "MiddleButton" then
		retVal = retVal..L["Middle click to complete"];
	else
		retVal = L["Open quest log to complete"];
	end
	return retVal;
end

function QWC:OnQuestLogChanged(...)
	-- reset reference list for updating
	--[[list.quest = nil;
	list.achievement = nil;
	list.quest = {};
	list.achievement = {};]]
	
	-- prepare fresh list of font strings
	fsList = fsList or {};
	for k, v in pairs(fsList) do
		fsList[k]:SetText("");
		fsList[k]:Hide();
	end
	
	-- prepare fresh list of popups
	pqList = pqList or {};
	for k, v in pairs(pqList) do
		pqList[k]:Hide();
	end
	
	-- prepare fresh list of buttons
	qiList = qiList or {};
	for k, v in pairs(qiList) do
		qiList[k]:Hide();
	end
	qbList = qbList or {};
	for k, v in pairs(qbList) do
		qbList[k]:Hide();
	end
	
	local numQuests = GetNumQuestLogEntries();
	
	-- font string, quest item, quest button, popup quest
	local fidx, qidx, bidx, pidx = 1, 1, 1, 1;
	
	local yOffset = self.db.profile["verticalInset"];
	
	-- Personal Note
	if QWC.db.profile["note"] == nil then
		if QWCNote then
			QWCNote:Hide();
			QWCNoteBody:Hide();
		end
	elseif QWC.db.profile["note"] == "clear" then
		QWC.db.profile["note"] = nil;
		if QWCNote then
			QWCNote:Hide();
			QWCNoteBody:Hide();
		end
	else
		local title, body = QWC:SplitNote();
		
		if QWCNote == nil then
			QWCNote = QWFrame:CreateFontString("QWCNote", "ARTWORK");
			QWCNoteBody = QWFrame:CreateFontString("QWCNoteBody", "ARTWORK");
		else
			QWCNote:Show();
			QWCNoteBody:Show();
		end
		
		QWCNote:SetFontObject("QWCTitle");
		QWCNoteBody:SetFontObject("QWCText");
		QWCNote:SetText(title);
		QWCNoteBody:SetText(body);
		QWCNote:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
		QWCNoteBody:SetWidth(QWFrame:GetWidth() - (self.db.profile["objectiveInset"] * 2));
		QWCNote:SetPoint("TOPLEFT", QWFrame, "TOPLEFT", self.db.profile["titleInset"], -yOffset);
		QWCNoteBody:SetPoint("TOPLEFT", QWCNote, "BOTTOMLEFT", self.db.profile["objectiveInset"], 0);
		
		if QWCNote:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
			QWCNote:Hide();
			return;
		end
		if QWCNoteBody:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
			QWCNoteBody:Hide();
			return;
		end
		
		yOffset = yOffset + QWCNote:GetHeight() + QWCNoteBody:GetHeight();
		
		-- button
		if self.db.profile["clickable"] then
			if qbList[bidx] == nil then
				qbList[bidx] = QWC:CreateButtonOverlay(bidx);
			else
				qbList[bidx]:Show();
			end
			qbList[bidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
			qbList[bidx]:SetHeight(QWCNote:GetHeight() + QWCNoteBody:GetHeight());
			qbList[bidx]:SetPoint("TOPLEFT", QWCNote, "TOPLEFT");
			qbList[bidx]:SetScript("OnClick", function(self, button)
				for action, func in pairs(noteEvents) do
					if button == QWC.db.profile[action] and QWC:Modifier(QWC.db.profile[action.."Modifier"]) then
						func();
						break;
					end
				end
			end);
			qbList[bidx]:SetScript("OnEnter", function()
				-- set highlight color on title and objectives
				QWCNote:SetFontObject("QWCTitleHighlight");
				QWCNoteBody:SetFontObject("QWCTextHighlight");
			end);
			qbList[bidx]:SetScript("OnLeave", function()
				-- back to normal color
				QWCNote:SetFontObject("QWCTitle");
				QWCNoteBody:SetFontObject("QWCText");
			end);
			
			bidx = bidx + 1;
		end
	end
	
	-- Check popup quests first, to keep them at the top of the frame.
	for questId, info in pairs(self.db.char["popupQuests"]) do
		if info ~= true and info.name == nil then
			local name, _, _, _, _, _, complete = GetQuestLogTitle(GetQuestLogIndexByID(questId));
			info.name = name;
			info.complete = complete;
		end
		if info ~= true and (info.name and info.name ~= "") then
			if pqList[pidx] == nil then
				pqList[pidx] = QWC:CreatePopup(pidx);
			else
				pqList[pidx]:Show();
			end
			pqList[pidx].index = pidx;
			
			if (info.complete and info.type == "COMPLETE") then
				pqList[pidx].ScrollChild.QuestionMark:Show();
				pqList[pidx].ScrollChild.Exclamation:Hide();
				pqList[pidx].ScrollChild.TopText:SetFontObject("QWCTitle");
				pqList[pidx].ScrollChild.TopText:SetText(L["Quest Completed!"]);
				pqList[pidx].ScrollChild.BottomText:SetFontObject("QWCGray");
				pqList[pidx].ScrollChild.BottomText:SetText(QUEST_WATCH_POPUP_CLICK_TO_COMPLETE);
				pqList[pidx].ScrollChild.Flash:Show();
				pqList[pidx].type="COMPLETED";
			elseif (info.type == "OFFER") then
				pqList[pidx].ScrollChild.QuestionMark:Hide();
				pqList[pidx].ScrollChild.Exclamation:Show();
				pqList[pidx].ScrollChild.TopText:SetFontObject("QWCTitle");
				pqList[pidx].ScrollChild.TopText:SetText(QUEST_WATCH_POPUP_QUEST_DISCOVERED);
				pqList[pidx].ScrollChild.BottomText:SetFontObject("QWCGray");
				pqList[pidx].ScrollChild.BottomText:SetText(QUEST_WATCH_POPUP_CLICK_TO_VIEW);
				pqList[pidx].ScrollChild.Flash:Show();
				pqList[pidx].type="OFFER";
			end
			
			if pidx == 1 then
				pqList[pidx]:SetPoint("TOPLEFT", QWFrame, "TOPLEFT", self.db.profile["titleInset"], -yOffset);
			else
				pqList[pidx]:SetPoint("TOPLEFT", pqList[pidx - 1], "BOTTOMLEFT", 0, 0);
			end
			
			pqList[pidx].ScrollChild.QuestName:SetFontObject("QWCText");
			pqList[pidx].ScrollChild.QuestName:SetText(info.name);
			local height = pqList[pidx].ScrollChild.TopText:GetStringHeight() + pqList[pidx].ScrollChild.QuestName:GetStringHeight() + pqList[pidx].ScrollChild.BottomText:GetStringHeight();
			pqList[pidx].ScrollChild:SetHeight(height);
			pqList[pidx]:SetHeight(height);
			pqList[pidx].ScrollChild:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
			pqList[pidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
			--pqList[pidx].ScrollChild:SetBackdrop(nil);
			pqList[pidx].questId = questId;
			
			-- mouseover highlights
			pqList[pidx]:SetScript("OnEnter", function(self)
				-- set highlight color
				self.ScrollChild.TopText:SetFontObject("QWCTitleHighlight");
				self.ScrollChild.QuestName:SetFontObject("QWCTextHighlight");
			end);
			pqList[pidx]:SetScript("OnLeave", function(self)
				-- back to normal color
				self.ScrollChild.TopText:SetFontObject("QWCTitle");
				self.ScrollChild.QuestName:SetFontObject("QWCText");
			end);
			
			-- scale texture
			local mark;
			if info.type == "COMPLETE" then
				mark = pqList[pidx].ScrollChild.QuestionMark;
			else
				mark = pqList[pidx].ScrollChild.Exclamation;
			end
			local tHeight = height * 0.8;
			local oldHeight = mark:GetHeight();
			local tWidth;
			if oldHeight ~= tHeight then
				local oldWidth = mark:GetWidth();
				tWidth = (tHeight / oldHeight) * oldWidth;
				mark:SetWidth(tWidth);
				mark:SetHeight(tHeight);
			end
			mark:SetPoint("TOPLEFT", pqList[pidx], "TOPLEFT", 0, -(height * 0.1));
			pqList[pidx].ScrollChild.TopText:SetPoint("TOPLEFT", pqList[pidx], "TOPLEFT", tWidth, 0);
			pqList[pidx].ScrollChild.QuestName:SetPoint("TOPLEFT", pqList[pidx].ScrollChild.TopText, "BOTTOMLEFT");
			pqList[pidx].ScrollChild.BottomText:SetPoint("TOPLEFT", pqList[pidx].ScrollChild.QuestName, "BOTTOMLEFT");
			
			-- highlight flash
			pqList[pidx].ScrollChild.Flash.Flash:SetHeight(pqList[pidx]:GetHeight());
			pqList[pidx].ScrollChild.Flash.Flash:SetWidth(pqList[pidx]:GetWidth());
			pqList[pidx].ScrollChild.Flash.Flash:ClearAllPoints();
			pqList[pidx].ScrollChild.Flash.Flash:SetPoint("TOPLEFT", pqList[pidx], "TOPLEFT", -20, 10);
			pqList[pidx].ScrollChild.Flash.Flash:SetPoint("BOTTOMRIGHT", pqList[pidx], "BOTTOMRIGHT", 10, -5);
			
			yOffset = yOffset + height;
			
			pidx = pidx + 1;
		end
	end
	
	-- Achievements
	for aID, info in pairsByKeys(self.db.char["trackedAchievements"]) do
		if info.complete ~= 1 then
			-- title
			if fsList[fidx] == nil then
				fsList[fidx] = QWC:CreateString(fidx, true);
			else
				fsList[fidx]:Show();
			end
			fsList[fidx]:SetFontObject("QWCTitle");
			fsList[fidx]:SetText(info.name);
			fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
			
			if fidx == 1 then
				fsList[fidx]:SetPoint("TOPLEFT", QWFrame, "TOPLEFT", self.db.profile["titleInset"], -yOffset);
			else
				if QWC:IsTitleFont(fsList[fidx - 1]:GetFontObject():GetName()) then
					fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", 0, 0);
				else
					fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", -self.db.profile["objectiveInset"], 0);
				end
			end
					
			if fsList[fidx]:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
				fsList[fidx]:Hide();
				return;
			end
			
			local entryHeight = fsList[fidx]:GetHeight();
			local entryIndex = fidx;
			
			list.achievement[aID] = fsList[fidx];
			list.achievement[aID].child = {};
			
			fidx = fidx + 1;
			
			if info.objCount == 0 then -- display description instead
				if fsList[fidx] == nil then
					fsList[fidx] = QWC:CreateString(fidx, false);
				else
					fsList[fidx]:Show();
				end
				fsList[fidx]:SetFontObject("QWCText");
				fsList[fidx]:SetText(info.desc);
				fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["objectiveInset"] * 2));
				
				fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", self.db.profile["objectiveInset"], 0);
				
				if fsList[fidx]:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
					fsList[fidx]:Hide();
					return;
				end
				
				entryHeight = entryHeight + fsList[fidx]:GetHeight();
				
				list.achievement[aID].child[1] = fsList[fidx];
				
				fidx = fidx + 1;
			else
				local first = true;
				-- objectives
				for idx = 1, info.objCount do
					if info["obj"..idx].objComplete ~= true then
						if fsList[fidx] == nil then
							fsList[fidx] = QWC:CreateString(fidx, false);
						else
							fsList[fidx]:Show();
						end
						fsList[fidx]:SetFontObject("QWCText");
						if info["obj"..idx].reqQuantity > 1 then -- append ": X/Y" to match quest layout
							if self.db.profile["textStyle"] == TS_NUMB then
								fsList[fidx]:SetText(info["obj"..idx].quantity.."/"..info["obj"..idx].reqQuantity.." "..info["obj"..idx].objDesc);
							elseif self.db.profile["textStyle"] == TS_DESC then
								fsList[fidx]:SetText(info["obj"..idx].objDesc..": "..info["obj"..idx].quantity.."/"..info["obj"..idx].reqQuantity);
							end
						else
							fsList[fidx]:SetText(info["obj"..idx].objDesc);
						end
						fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2) - (self.db.profile["objectiveInset"] * 2));
						
						if first == true then
							fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", self.db.profile["objectiveInset"], 0);
						else
							fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT");
						end
						
						if fsList[fidx]:GetBottom() < QWFrame:GetBottom() + (self.db.profile["verticalInset"]) then
							fsList[fidx]:Hide();
							return;
						end
						
						entryHeight = entryHeight + fsList[fidx]:GetHeight();
				
						list.achievement[aID].child[idx] = fsList[fidx];
						
						fidx = fidx + 1;
					end
					if info["obj"..idx].objComplete ~= true and first == true then
						first = false;
					end
				end
			end
			
			-- clickable button
			if self.db.profile["clickable"] then
				if qbList[bidx] == nil then
					qbList[bidx] = QWC:CreateButtonOverlay(bidx);
				else
					qbList[bidx]:Show();
				end
				qbList[bidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
				qbList[bidx]:SetHeight(entryHeight);
				qbList[bidx]:SetPoint("TOPLEFT", fsList[entryIndex], "TOPLEFT");
				qbList[bidx]:SetScript("OnClick", function(self, button)
					for action, func in pairs(achievementEvents) do
						if button == QWC.db.profile[action] and QWC:Modifier(QWC.db.profile[action.."Modifier"]) then
							func(aID);
							break;
						end
					end
				end);
				qbList[bidx]:SetScript("OnEnter", function()
					-- set highlight color on title and objectives
					fsList[entryIndex]:SetFontObject("QWCTitleHighlight");
					local idx = 1;
					for jdx = 1, info.objCount do
						if info["obj"..jdx].objComplete ~= true then
							fsList[idx + entryIndex]:SetFontObject("QWCTextHighlight");
							idx = idx + 1;
						end
					end
				end);
				qbList[bidx]:SetScript("OnLeave", function()
					-- back to normal color
					fsList[entryIndex]:SetFontObject("QWCTitle");
					local idx = 1;
					for jdx = 1, info.objCount do
						if info["obj"..jdx].objComplete ~= true then
							fsList[idx + entryIndex]:SetFontObject("QWCText");
							idx = idx + 1;
						end
					end
				end);
				
				bidx = bidx + 1;
			end
		end
	end
	
	-- quests
	for idx = 1, numQuests, 1 do
		SelectQuestLogEntry(idx);
		local numBoards = GetNumQuestLeaderBoards(idx);
		if IsQuestWatched(idx) ~= nil then
			local title, level, _, _, _, _, complete, _, questID = GetQuestLogTitle(idx);
			--if complete ~= 1 then
			-- Quest title string
			if fsList[fidx] == nil then
				fsList[fidx] = QWC:CreateString(fidx, true);
			else
				fsList[fidx]:Show();
			end
			if QWC.db.profile["textColor"] == TC_USER then
				fsList[fidx]:SetFontObject("QWCTitle");
			elseif QWC.db.profile["textColor"] == TC_DIFF then
				fsList[fidx]:SetFontObject(QWC:GetQuestDiffFont(level));
			end
			fsList[fidx]:SetText(title);
			fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
			
			if fidx == 1 then
				fsList[fidx]:SetPoint("TOPLEFT", QWFrame, "TOPLEFT", self.db.profile["titleInset"], -yOffset);
			else
				if QWC:IsTitleFont(fsList[fidx - 1]:GetFontObject():GetName()) then
					fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", 0, 0);
				else
					fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", -self.db.profile["objectiveInset"], 0);
				end
			end
			
			if fsList[fidx]:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
				fsList[fidx]:Hide();
				return;
			end
			
			local entryHeight = fsList[fidx]:GetHeight();
			local entryIndex = fidx;
				
			list.quest[idx] = fsList[fidx];
			list.quest[idx].child = {};
			
			fidx = fidx + 1;
			
			local showQuestItem = false;
			
			-- objectives
			local first = true;
			for jdx = 1, numBoards, 1 do
				local info, qType, done = GetQuestLogLeaderBoard(jdx, idx);
				if info ~= nil and done ~= 1 or qType == "log" then
					-- Quest objective strings
					if fsList[fidx] == nil then
						fsList[fidx] = QWC:CreateString(fidx, false);
					else
						fsList[fidx]:Show();
					end
					if QWC.db.profile["textObjColor"] == TC_USER then
						fsList[fidx]:SetFontObject("QWCText");
					elseif QWC.db.profile["textObjColor"] == TC_DIFF then
						fsList[fidx]:SetFontObject(QWC:GetQuestDiffFont(level, true));
					end
					if self.db.profile["textStyle"] == TS_NUMB then
						local newInfo = QWC:Split(info, ": ");
						local fixedDesc = "";
						for kdx = 1, (#newInfo - 1) do
							fixedDesc = fixedDesc..newInfo[kdx];
						end
						fsList[fidx]:SetText(newInfo[#newInfo].." "..fixedDesc);
					elseif self.db.profile["textStyle"] == TS_DESC then
						fsList[fidx]:SetText(info);
					end
					fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2) - (self.db.profile["objectiveInset"] * 2));
					
					if first == true then
						fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", self.db.profile["objectiveInset"], 0);
					else
						fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT");
					end
					
					if fsList[fidx]:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
						fsList[fidx]:Hide();
						return;
					end
					
					entryHeight = entryHeight + fsList[fidx]:GetHeight();
				
					list.quest[idx].child[jdx] = fsList[fidx];
					
					fidx = fidx + 1;
					
					if showQuestItem == false then
						showQuestItem = true;
					end
				end
				if done ~= 1 and first == true then
					first = false;
				end
			end
			--end
			
			-- description if no objectives
			local qDesc;
			local objText = nil;
			if numBoards == nil or numBoards == 0 then
				qDesc_, objText = GetQuestLogQuestText();
				if fsList[fidx] == nil then
					fsList[fidx] = QWC:CreateString(fidx, false);
				else
					fsList[fidx]:Show();
				end
				if QWC.db.profile["textObjColor"] == TC_USER then
					fsList[fidx]:SetFontObject("QWCText");
				elseif QWC.db.profile["textObjColor"] == TC_DIFF then
					fsList[fidx]:SetFontObject(QWC:GetQuestDiffFont(level, true));
				end
				fsList[fidx]:SetText(objText);
				fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2) - (self.db.profile["objectiveInset"] * 2));
				
				fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", self.db.profile["objectiveInset"], 0);
				
				if fsList[fidx]:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
					fsList[fidx]:Hide();
					return;
				end
				
				entryHeight = entryHeight + fsList[fidx]:GetHeight();
				
				list.quest[idx].child[1] = fsList[fidx];
				
				fidx = fidx + 1;
					
				if showQuestItem == false then
					showQuestItem = true;
				end
			end
			
			-- if completed quest popup, display "click to complete"
			if self.db.char["popupQuests"][questID] == true then
				if fsList[fidx] == nil then
					fsList[fidx] = QWC:CreateString(fidx, false);
				else
					fsList[fidx]:Show();
				end
				if QWC.db.profile["textObjColor"] == TC_USER then
					fsList[fidx]:SetFontObject("QWCText");
				elseif QWC.db.profile["textObjColor"] == TC_DIFF then
					fsList[fidx]:SetFontObject(QWC:GetQuestDiffFont(level, true));
				end
				fsList[fidx]:SetText(QWC:GetLogClickText());
				fsList[fidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2) - (self.db.profile["objectiveInset"] * 2));
				
				fsList[fidx]:SetPoint("TOPLEFT", fsList[fidx - 1], "BOTTOMLEFT", self.db.profile["objectiveInset"], 0);
				
				if fsList[fidx]:GetBottom() < (QWFrame:GetBottom() + self.db.profile["verticalInset"]) then
					fsList[fidx]:Hide();
					return;
				end
				
				entryHeight = entryHeight + fsList[fidx]:GetHeight();
				
				list.quest[idx].child[1] = fsList[fidx];
				
				fidx = fidx + 1;
			end
			
			-- Quest item button
			local iLink, iIcon, iCharges = GetQuestLogSpecialItemInfo(idx);
			if showQuestItem and iLink ~= nil then -- we have a quest item
				if qiList[qidx] == nil then
					qiList[qidx] = QWC:CreateButton(qidx);
				else
					qiList[qidx]:Show();
				end
				qiList[qidx].ItemID = iLink:match("Hitem:(%d+):");
				
				qiList[qidx]:SetID(idx);
				
				SetItemButtonTexture(qiList[qidx], iIcon);
				SetItemButtonTextureVertexColor(qiList[qidx], 1.0, 1.0, 1.0);
				SetItemButtonCount(qiList[qidx], iCharges);
				WatchFrameItem_UpdateCooldown(qiList[qidx]);
				_G[qiList[qidx]:GetName().."HotKey"]:Hide();
				
				qiList[qidx].rangeTimer = -1;
				qiList[qidx]:RegisterEvent("BAG_UPDATE_COOLDOWN")
				
				if (ItemHasRange(iLink) ~= nil) then
					qiList[qidx]:RegisterEvent("PLAYER_TARGET_CHANGED")
					qiList[qidx]:SetScript("OnUpdate", function(self, elapsed)
						self.rangeTimer = self.rangeTimer - elapsed
						if self.rangeTimer <= 0 then
							if IsQuestLogSpecialItemInRange(self:GetID()) == 0 then
								SetItemButtonTextureVertexColor(self, 0.8, 0.1, 0.1)
							else
								SetItemButtonTextureVertexColor(self, 1.0, 1.0, 1.0)
							end
							self.rangeTimer = TOOLTIP_UPDATE_TIME
						end
					end)
				end
				
				qiList[qidx]:SetScript("OnEvent", function(self, event)
					if (event == "PLAYER_TARGET_CHANGED") then
						self.rangeTimer = -1
					elseif (event == "BAG_UPDATE_COOLDOWN") then
						local objCooldown = _G[self:GetName().."Cooldown"]
						local startTime, duration, enable = GetItemCooldown(self.ItemID)
						if startTime > 0 and duration > 0 and enable > 0 then
							objCooldown:SetCooldown(startTime, duration)
							objCooldown:Show()
						else
							objCooldown:Hide()
						end
					end
				end)
				
				-- set scale to title height + objective height
				qiList[qidx]:SetScale((QWC.db.profile["titleSize"] + QWC.db.profile["objectiveSize"]) / qiList[qidx]:GetHeight());
				
				qiList[qidx]:SetPoint("TOPRIGHT", fsList[entryIndex], "TOPLEFT", -5 - self.db.profile["titleInset"], 0);
				
				qidx = qidx + 1;
			end
			
			-- button overlay
			if self.db.profile["clickable"] then
				if qbList[bidx] == nil then
					qbList[bidx] = QWC:CreateButtonOverlay(bidx);
				else
					qbList[bidx]:Show();
				end
				qbList[bidx]:SetWidth(QWFrame:GetWidth() - (self.db.profile["titleInset"] * 2));
				qbList[bidx]:SetHeight(entryHeight);
				qbList[bidx]:SetPoint("TOPLEFT", fsList[entryIndex], "TOPLEFT");
				qbList[bidx]:SetScript("OnClick", function(self, button)
					for action, func in pairs(questEvents) do
						if button == QWC.db.profile[action] and QWC:Modifier(QWC.db.profile[action.."Modifier"]) then
							func(idx, questID);
							break;
						end
					end
				end);
				qbList[bidx]:SetScript("OnEnter", function()
					-- set highlight color on title and objectives
					if QWC.db.profile["textColor"] == TC_USER then
						fsList[entryIndex]:SetFontObject("QWCTitleHighlight");
					elseif QWC.db.profile["textColor"] == TC_DIFF then
						fsList[entryIndex]:SetFontObject(QWC:GetQuestDiffFontHighlight(level));
					end
					local kdx = 1;
					for jdx = 1, numBoards do
						local info, qType, done = GetQuestLogLeaderBoard(jdx, idx);
						if info ~= nil and done ~= 1 or qType == "log" then
							if QWC.db.profile["textObjColor"] == TC_USER then
								fsList[kdx + entryIndex]:SetFontObject("QWCTextHighlight");
							elseif QWC.db.profile["textObjColor"] == TC_DIFF then
								fsList[kdx + entryIndex]:SetFontObject(QWC:GetQuestDiffFontHighlight(level, true));
							end
							kdx = kdx + 1;
						end
					end
					if objText ~= nil then
						if QWC.db.profile["textObjColor"] == TC_USER then
							fsList[kdx + entryIndex]:SetFontObject("QWCTextHighlight");
						elseif QWC.db.profile["textObjColor"] == TC_DIFF then
							fsList[kdx + entryIndex]:SetFontObject(QWC:GetQuestDiffFontHighlight(level, true));
						end
					end
				end);
				qbList[bidx]:SetScript("OnLeave", function()
					-- back to normal color
					if QWC.db.profile["textColor"] == TC_USER then
						fsList[entryIndex]:SetFontObject("QWCTitle");
					elseif QWC.db.profile["textColor"] == TC_DIFF then
						fsList[entryIndex]:SetFontObject(QWC:GetQuestDiffFont(level));
					end
					local kdx = 1;
					for jdx = 1, numBoards do
						local info, qType, done = GetQuestLogLeaderBoard(jdx, idx);
						if info ~= nil and done ~= 1 or qType == "log" then
							if QWC.db.profile["textObjColor"] == TC_USER then
								fsList[kdx + entryIndex]:SetFontObject("QWCText");
							elseif QWC.db.profile["textObjColor"] == TC_DIFF then
								fsList[kdx + entryIndex]:SetFontObject(QWC:GetQuestDiffFont(level, true));
							end
							kdx = kdx + 1;
						end
					end
					if objText ~= nil then
						if QWC.db.profile["textObjColor"] == TC_USER then
							fsList[kdx + entryIndex]:SetFontObject("QWCText");
						elseif QWC.db.profile["textObjColor"] == TC_DIFF then
							fsList[kdx + entryIndex]:SetFontObject(QWC:GetQuestDiffFont(level, true));
						end
					end
				end);
				
				bidx = bidx + 1;
			end
		end
	end
end

function QWC:CreateString(index, isTitle)
	local retVal = fsList[index] or QWFrame:CreateFontString("QWCFS"..index, "ARTWORK");
	--[[if isTitle then
		retVal:SetWidth(QWFrame:GetWidth() - 10);
	else
		retVal:SetWidth(QWFrame:GetWidth() - 20);
	end]]
	return retVal;
end

function QWC:CreatePopup(index)
	local retVal = CreateFrame("SCROLLFRAME", "QWCPQ"..index, UIParent, "QWCPopupTemplate"--[["WatchFrameAutoQuestPopUpTemplate"]]);
	return retVal;
end

function QWC:CreateButton(index)
	local retVal = CreateFrame("BUTTON", "QWCQI"..index, UIParent, "WatchFrameItemButtonTemplate");
	retVal:SetScale(1.0);
	return retVal;
end

function QWC:CreateButtonOverlay(index)
	local retVal = CreateFrame("BUTTON", "QWCQB"..index, QWFrame);
	retVal:RegisterForClicks("AnyUp");
	return retVal;
end

function QWC:Show()
	ShowUIPanel(QWFrame);
	for event, method in pairs(events) do
		QWC:UnregisterEvent(event);
		QWC:RegisterEvent(event, method);
	end
	for hook, method in pairs(hooks) do
		QWC:Unhook(hook);
		QWC:SecureHook(hook, method);
	end
	QWC:OnQuestLogChanged();
end

function QWC:Hide()
	HideUIPanel(QWFrame);
	for event, method in pairs(events) do
		QWC:UnregisterEvent(event);
	end
	for hook, method in pairs(hooks) do
		QWC:Unhook(hook);
	end
	qiList = qiList or {};
	for k, v in pairs(qiList) do
		qiList[k]:Hide();
	end
end